/*
 * Copyright (C) Bigly Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

package com.biglybt.pifimpl.local.ui.components;

import com.biglybt.pif.ui.components.UIButton;

public class 
UIButtonImpl
	extends		UIComponentImpl
	implements 	UIButton
{
	private String 		name;
	private String 		label;
	
	@Override
	public void 
	setLabel(
		String	_label )
	{
		label		= _label;
	}
	
	@Override
	public String 
	getLabel()
	{
		return( label );
	}
	
	@Override
	public void 
	setName(
		String	_name)
	{
		name		= _name;
	}
	
	@Override
	public String 
	getName()
	{
		return( name );
	}
}
