/*
 * Created : 10 oct. 2003 16:07:53
 * By      : Olivier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details ( see the LICENSE file ).
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.biglybt.ui.swt.config;

import org.eclipse.swt.widgets.Control;

public interface SwtParameter<VALUETYPE>
{
	boolean DEBUG = false;

	/** Parameter has changed.  Should probably refreshControl any UI components */
	void informChanged();

	void setLayoutData(Object layoutData);

	Control getMainControl();

	Control[] getControls();

	void setEnabled(boolean enabled);

	boolean isDisposed();

	void setValue(VALUETYPE value);

	VALUETYPE getValue();

	String getParamID();

	void setVisible(boolean visible);

	void debug(String s);
}
