/*
 * Copyright (C) Azureus Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

package com.biglybt.core.util;

/**
 * A Runnable that calls the AECallback function when it's done running the
 * code the implementer supplied
 *
 * @author TuxPaper
 * @created Mar 22, 2007
 *
 */
public abstract class AERunnableWithCallback
	implements Runnable
{
	private final AECallback callback;

	public AERunnableWithCallback(AECallback callback) {
		this.callback = callback;
	}

	@Override
	public final void run() {
		try {
			Object o = runSupport();
			if (callback != null) {
				callback.callbackSuccess(o);
			}
		} catch (Throwable e) {
			Debug.out(e);
			if (callback != null) {
				callback.callbackFailure(e);
			}
		}
	}

	public abstract Object runSupport();
}
