/*
 * Copyright (C) Bigly Software.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.biglybt.ui.swt.config;

import com.biglybt.pif.ui.config.ParameterValidator.ValidationInfo;

/**
 * Pretty much @{@link com.biglybt.pif.ui.config.ParameterValidator},
 * except isValidParameterValue is cast to the correct SwtParameter
 */
public interface SwtParameterValidator<PARAMTYPE extends SwtParameter<VALUETYPE>, VALUETYPE>
{
	/**
	 * Determine validility of parameter value.  Called before permanently
	 * setting config value, as well as when user is editing field
	 */
	ValidationInfo isValidParameterValue(PARAMTYPE p, VALUETYPE toValue);
}
