/*
 * UITemplate.java
 *
 * Created on 26. Oktober 2003, 22:40
 * Copyright (C) Azureus Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.biglybt.ui.common;

import com.biglybt.core.Core;
import org.apache.commons.cli.CommandLine;

/**
 *
 * @author  tobi
 */
public abstract class
UITemplate
	implements IUserInterface
{

  protected Core core;
  private boolean first;
  private boolean others;

  /** Creates a new instance of UITemplate */
  public UITemplate() {
  }

  @Override
  public void takeMainThread() {}

  @Override
  public void coreCreated(Core core) {
    this.core = core;
  }

  @Override
  public void init(boolean first, boolean others) {
    this.first = first;
    this.others = others;
  }

  @Override
  abstract public String[] processArgs(CommandLine commands, String[] args);

  public boolean isFirst() {
    return first;
  }

  public boolean hasOthers() {
    return others;
  }
}
