/*
 * Copyright (C) Azureus Software, Inc, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details ( see the LICENSE file ).
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.biglybt.ui.common;

import com.biglybt.pif.ui.toolbar.UIToolBarActivationListener;
import com.biglybt.pif.ui.toolbar.UIToolBarItem;

public interface ToolBarItem
	extends UIToolBarItem
{
	public static interface ToolBarItemListener
	{
		public void uiFieldChanged(ToolBarItem item);

		public boolean triggerToolBarItem(ToolBarItem item, long activationType,
				Object datasource);
	}

	@Override
	public boolean triggerToolBarItem(long activationType, Object datasource);

	@Override
	public void setDefaultActivationListener(
			UIToolBarActivationListener toolBarActivation);

	public void setAlwaysAvailable(boolean b);

	public UIToolBarActivationListener getDefaultActivationListener();

	public void addToolBarItemListener(ToolBarItemListener l);

	public void removeToolBarItemListener(ToolBarItemListener l);
}
